#
# Weather-related plugin services.
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time

import asyn
import forecast
import astro

import cyin.devstate
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *


MIN_REFRESH = 5		# enforced minimum minutes between update calls per location


#
# Common darksky.net snapshot data device
#
class ForecastDevice(cyin.devstate.Device):

	summary = cyin.DeviceState(type=str)
	dewPoint = cyin.DeviceState(type=float)
	ozone = cyin.DeviceState(type=float)
	pressure = cyin.DeviceState(type=float, format=" mb")
	humidity = cyin.DeviceState(type=float, format="%")
	visibility = cyin.DeviceState(type=float, format=" mi")
	cloudCover = cyin.DeviceState(type=float, format="%")
	windSpeed = cyin.DeviceState(type=float, format=" mph")
	windBearing = cyin.DeviceState(type=float, format="\xb0")
	nearestStormDistance = cyin.DeviceState(type=float, format=" mi")
	nearestStormBearing = cyin.DeviceState(type=float, format="\xb0")
	icon = cyin.DeviceState(type=str)
	data = cyin.DeviceState(type=str)

	def updateReading(self, reading):
		def update(name, op=lambda s: s):
			value = getattr(reading, name)
			if value is not None:
				setattr(self, name, op(value))
		update("summary")
		update("temperature", lambda s: round(s, 1))
		update("apparentTemperature", lambda s: round(s, 1))
		update("temperatureMin", lambda s: round(s, 1))
		update("temperatureMax", lambda s: round(s, 1))
		update("apparentTemperatureMin", lambda s: round(s, 1))
		update("apparentTemperatureMax", lambda s: round(s, 1))
		update("dewPoint", lambda s: round(s, 1))
		update("ozone")
		update("pressure")
		update("humidity", lambda s: round(s*100))
		update("visibility")
		update("cloudCover", lambda s: round(s*100))
		update("windSpeed")
		update("windBearing")
		update("nearestStormDistance")
		update("nearestStormBearing")
		update("icon")
		update("moonPhase")


#
# A geographic location collecting and holding a full Dark Sky data set.
# Its Indigo properties represent present-time values.
#
class Location(ForecastDevice):
	config_version = 2
	
	latitude = cyin.PluginProperty(type=float, required=False, check=[check_range(-90,90)])
	longitude = cyin.PluginProperty(type=float, required=False, check=[check_range(-180,180)])
	polling = cyin.PluginProperty(type=int, required=False, check=[check_range(min=MIN_REFRESH)])
	units = cyin.PluginProperty(type=str)
	rawdata = cyin.PluginProperty(type=bool)
	
	alert = cyin.DeviceState(type=str)
	alert_url = cyin.DeviceState(type=str)
	temperature = cyin.DeviceState(type=float)
	apparentTemperature = cyin.DeviceState(type=float)
	moonPhase = cyin.DeviceState(type=float)

	_poll_timer = None
	_last_update = 0
	lastReading = None

	def start(self):
		super(Location, self).start()
		self.setup()

	def setup(self, ctx=None):
		self.forecast = forecast.Forecast(cyin.plugin)
		self.forecast.apikey = cyin.plugin.apikey
		self.forecast.location = self._location()
		self.forecast.units = self.units
		self.forecast.user_agent = 'Cynical-Weather/%s' % self.plugin.version
		self.set_display_address(astro.s_location(*self.forecast.location.lat_lon))
		if self.polling:
			self.update()
		else:
			self.proceed("ready")	# (passive mode)

	def stop(self):
		if self._poll_timer:
			self._poll_timer.cancel()
			self._poll_timer = None
		super(Location, self).stop()

	def update(self, ctx=None):
		""" Manage automatic timed polling updates. """
		if self.polling:
			self._poll_timer = cyin.plugin.schedule(self.update, after=self.polling * 60)
		self.poll(force=not ctx)

	def poll(self, force=False):
		""" Explicitly poll for new data from darksky.net about this location.

			This is clamped to no more than once per MIN_REFRESH minutes to avoid
			accidentally hitting the daily request limit.
		"""
		now = time.time()
		if not force and self._last_update + MIN_REFRESH * 60 > now:
			return error(self.name, "ignoring update request within %d minutes of last one" % MIN_REFRESH)
		self._last_update = now
		def updated(ctx, data=None):
			if ctx.error:
				return self.fail_hard(ctx)
			elif ctx.state == 'error':
				return self.fail_hard("darksky.net service error: %s %s" % (data.n_status, data.v_status))
			elif ctx.state == 'reading':
				self.lastReading = data
				debug(self.name, "updated")
				self.data = data.raw if self.rawdata else "N/A"
				self.moonPhase = data.days[0].moonPhase		# missing in current data
				self.updateAlerts(data.alerts)
				self.updateReading(data.current)
				for fc in Forecast.all():
					fc.updateForecast(data)
				self.proceed("ready", recovered=True)
		self.forecast.poll(callout=updated)
	
	def updateAlerts(self, alerts):
		now = time.time()
		alerts = [alert for alert in alerts if alert.expires >= now]
		self.alert = '; '.join([alert.title for alert in alerts])
		self.alert_url = '; '.join([alert.uri for alert in alerts])

	def _location(self):
		""" Return a forecast.Location object from either explicit data or the default location. """
		if self.latitude and self.longitude:
			return forecast.Location(self.latitude, self.longitude)
		else:
			return forecast.Location(*indigo.server.getLatitudeAndLongitude())

	@cyin.action
	def update_data(self, action):
		self.poll()


#
# A temporal slice of an existing darksky.net data set.
#
class Forecast(ForecastDevice):
	location = cyin.PluginProperty(type=cyin.device)
	distance = cyin.PluginProperty(type=int, check=[check_range(0, 50)])

	def start(self):
		super(Forecast, self).start()
		self.set_hostdev(self.location)
		self.setup()

	def setup(self, ctx=None):
		self.set_display_address("%s + %d %s" % (
			self.hostdev.name, self.distance, self.units))
		if self.hostdev.lastReading:
			self.updateForecast(self.hostdev.lastReading)
		self.proceed("ready")

	def updateForecast(self, reading):
		data = reading.getPoint(self.distance, self.units)
		if data:
			debug(self.name, "updated for", time.ctime(data.time))
			self.updateReading(data)
		else:
			error("%s: %s has no reading for %d %s ahead" % (self.name, self.hostdev.name, self.distance, self.units))

	@cyin.action
	def update_data(self, action):
		self.hostdev.poll()


class HourForecast(Forecast):
	units = 'hours'
	
	temperature = cyin.DeviceState(type=float)
	apparentTemperature = cyin.DeviceState(type=float)


class DayForecast(Forecast):
	config_version = 2
	
	units = 'days'

	temperatureMin = cyin.DeviceState(type=float)
	temperatureMax = cyin.DeviceState(type=float)
	apparentTemperatureMin = cyin.DeviceState(type=float)
	apparentTemperatureMax = cyin.DeviceState(type=float)
	moonPhase = cyin.DeviceState(type=float)


#
# A (window et al) orientation in space. This is what we orient the Sun against.
# Actual updates are driven by a central timer.
#
BELOW_HORIZON = -5			# degrees below horizon to be effectively dark
FULL_HEIGHT = 15			# height above horizon to presume full solar impact

class Orientation(cyin.Device):
	""" An orientation in space (a "facing").

		This object has state indicating the direction TO the Sun for someone
		facing this direction. It pays no attention to occlusion or obstacles.
	"""
	facing = cyin.PluginProperty(type=int, check=[check_range(min=0, max=360)])

	azimuth = cyin.DeviceState(type=float, format="\xb0")
	height = cyin.DeviceState(type=float, format="\xb0")
	index = cyin.DeviceState(type=int)

	_refresh_timer = None

	def start(self):
		self.set_display_address(astro.s_bearing(self.facing))
		if cyin.plugin.active:			# runtime start
			cyin.plugin.updateSun()

	def _update(self, sun):
		(A, h) = sun		# (Azimuth to true S, angular height)

		#
		# A is south-relative (0 = due South, -90 = due East, +145 = Northwest)
		#
		az = A - self.facing + 180		# north-by-north delta 
		az = (az + 180) % 360 - 180		# normalize to [-180, +180]

		# straight angles
		self.azimuth = round(az, 2)		# sun azimuth relative to facing
		self.height = round(h, 2)		# sun height over horizontal

		# judgment calls (rudimentary)
		if h < 0 or abs(az) > 90:		# sun not shining on this surface
			self.index = 0
		else:
			az_index = 100 - abs(az) * 100.0 / 90	# azimuth scale [0..100] <-> [90..0]
			h_index = h * 100.0 / FULL_HEIGHT	# height scale [0..100] <-> [0..FULL_HEIGHT]
			self.index = min(az_index, h_index)


#
# Actions
#
class Poll(cyin.DeviceAction):
	pass


#
# Events
#
class Precipitation(cyin.Event):
	location = cyin.PluginProperty(type=cyin.device)
	threshold = cyin.PluginProperty(type=int, check=[check_int(min=10,max=100)])

	def matches(self, location):
		if self.location:
			if self.location != location:
				return False
#		now = time.time()
#		if self.last_motion and self.last_motion + self.throttle > now:
#			return False                # too close since last motion
#		self.last_motion = now          # set mark
		return True


#
# The plugin.
#
SUN_REFRESH = 1 * 60	# every minute
#SUN_REFRESH = 1		# debug (every second)

class Plugin(cyin.asynplugin.Plugin):

	apikey = cyin.PluginPreference(type=str, required=False, check=[check_format('[0-9a-fA-F]{32,200}', error="That doesn't look like a valid API key.")])

	def startup(self):
		cyin.asynplugin.Plugin.startup(self)
		self.setLocation()
		self._solar_refresh = self.schedule(self.updateSun)

	# update location on wakeup, just in case
	def wakeup(self):
		super(Plugin, self).wakeup()
		self.setLocation()
		self.updateSun()

	def setLocation(self):
		# set sun-tracking Location for "here"
		self._sunLoc = astro.SunLocation(*indigo.server.getLatitudeAndLongitude())
#		debug("location is", self._sunLoc)

	def updateSun(self, ctx=None):
		if ctx:
			ctx.reschedule(after=SUN_REFRESH)
		# else just do an update pass and leave the timer running
		sun = None
		for orient in Orientation.all():
			if sun is None:	# lazy
				sun = astro.sun_position(time.time(), self._sunLoc)
#				debug("sun is at", sun)
			orient._update(sun)
